@@ -11,11 +11,11 @@
  * and its documentation for any purpose is hereby granted without
  * fee, provided that the above copyright notice appear in all copies
  * and that both that copyright notice and this permission notice
- * appear in supporting documentation, and that the name of Red Hat
- * not be used in advertising or publicity pertaining to distribution
- * of the software without specific, written prior permission.  Red
- * Hat makes no representations about the suitability of this software
- * for any purpose.  It is provided "as is" without express or implied
+ * appear in supporting documentation, and that the name of the authors
+ * not be used in advertising or publicity pertaining to distribution of the
+ * software without specific, written prior permission.  The authors make no
+ * representations about the suitability of this software for any
+ * purpose.  It is provided "as is" without express or implied
  * warranty.
  *
  * THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
@@ -34,8 +34,12 @@
 #include "config.h"
 #endif
 
+#include <X11/Xatom.h>
 #include <xf86.h>
+#include <xf86Xinput.h>
+#include <exevents.h>
 
+#include <evdev-properties.h>
 #include "evdev.h"
 
 enum {
@@ -44,6 +48,10 @@ enum {
     MBEMU_AUTO
 };
 
+#ifdef HAVE_PROPERTIES
+static Atom prop_mbemu     = 0; /* Middle button emulation on/off property */
+static Atom prop_mbtimeout = 0; /* Middle button timeout property */
+#endif
 /*
  * Lets create a simple finite-state machine for 3 button emulation:
  *
@@ -206,14 +214,14 @@ EvdevMBEmuTimer(InputInfoPtr pInfo)
 /**
  * Emulate a middle button on button press.
  *
- * @param code Evdev event code (BTN_LEFT or BTN_RIGHT)
+ * @param code button number (1 for left, 3 for right)
  * @param press TRUE if press, FALSE if release.
  *
  * @return TRUE if event was swallowed by middle mouse button emulation, FALSE
  * otherwise.
  */
 BOOL
-EvdevMBEmuFilterEvent(InputInfoPtr pInfo, int code, BOOL press)
+EvdevMBEmuFilterEvent(InputInfoPtr pInfo, int button, BOOL press)
 {
     EvdevPtr pEvdev = pInfo->private;
     int id;
@@ -223,15 +231,20 @@ EvdevMBEmuFilterEvent(InputInfoPtr pInfo, int code, BOOL press)
     if (!pEvdev->emulateMB.enabled)
         return ret;
 
+    if (button == 2) {
+        EvdevMBEmuEnable(pInfo, FALSE);
+        return ret;
+    }
+
     /* don't care about other buttons */
-    if (code != BTN_LEFT && code != BTN_RIGHT)
+    if (button != 1 && button != 3)
         return ret;
 
     btstate = &pEvdev->emulateMB.buttonstate;
     if (press)
-        *btstate |= (code == BTN_LEFT) ? 0x1 : 0x2;
+        *btstate |= (button == 1) ? 0x1 : 0x2;
     else
-        *btstate &= (code == BTN_LEFT) ? ~0x1 : ~0x2;
+        *btstate &= (button == 1) ? ~0x1 : ~0x2;
 
     if ((id = stateTab[pEvdev->emulateMB.state][*btstate][0]) != 0)
     {
@@ -303,21 +316,31 @@ EvdevMBEmuPreInit(InputInfoPtr pInfo)
         pEvdev->emulateMB.enabled = xf86SetBoolOption(pInfo->options,
                                                       "Emulate3Buttons",
                                                       MBEMU_ENABLED);
-        xf86Msg(X_INFO, "%s: Forcing middle mouse button emulation.\n",
-                        pInfo->name);
+        xf86Msg(X_INFO, "%s: Forcing middle mouse button emulation %s.\n",
+                pInfo->name, (pEvdev->emulateMB.enabled) ? "on" : "off");
     }
 
     pEvdev->emulateMB.timeout = xf86SetIntOption(pInfo->options,
                                                  "Emulate3Timeout", 50);
+}
+
+void
+EvdevMBEmuOn(InputInfoPtr pInfo)
+{
+    if (!pInfo->dev->button) /* don't init for keyboards */
+        return;
+
     RegisterBlockAndWakeupHandlers (EvdevMBEmuBlockHandler,
                                     EvdevMBEmuWakeupHandler,
                                     (pointer)pInfo);
-
 }
 
 void
 EvdevMBEmuFinalize(InputInfoPtr pInfo)
 {
+    if (!pInfo->dev->button) /* don't cleanup for keyboards */
+        return;
+
     RemoveBlockAndWakeupHandlers (EvdevMBEmuBlockHandler,
                                   EvdevMBEmuWakeupHandler,
                                   (pointer)pInfo);
@@ -332,3 +355,67 @@ EvdevMBEmuEnable(InputInfoPtr pInfo, BOOL enable)
     if (pEvdev->emulateMB.enabled == MBEMU_AUTO)
         pEvdev->emulateMB.enabled = enable;
 }
+
+
+#ifdef HAVE_PROPERTIES
+static int
+EvdevMBEmuSetProperty(DeviceIntPtr dev, Atom atom, XIPropertyValuePtr val,
+                      BOOL checkonly)
+{
+    InputInfoPtr pInfo  = dev->public.devicePrivate;
+    EvdevPtr     pEvdev = pInfo->private;
+
+    if (atom == prop_mbemu)
+    {
+        if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
+            return BadMatch;
+
+        if (!checkonly)
+            pEvdev->emulateMB.enabled = *((BOOL*)val->data);
+    } else if (atom == prop_mbtimeout)
+    {
+        if (val->format != 32 || val->size != 1 || val->type != XA_INTEGER)
+            return BadMatch;
+
+        if (!checkonly)
+            pEvdev->emulateMB.timeout = *((CARD32*)val->data);
+    }
+
+    return Success;
+}
+
+/**
+ * Initialise property for MB emulation on/off.
+ */
+void
+EvdevMBEmuInitProperty(DeviceIntPtr dev)
+{
+    InputInfoPtr pInfo  = dev->public.devicePrivate;
+    EvdevPtr     pEvdev = pInfo->private;
+    int          rc;
+
+    if (!dev->button) /* don't init prop for keyboards */
+        return;
+
+    prop_mbemu = MakeAtom(EVDEV_PROP_MIDBUTTON, strlen(EVDEV_PROP_MIDBUTTON), TRUE);
+    rc = XIChangeDeviceProperty(dev, prop_mbemu, XA_INTEGER, 8,
+                                PropModeReplace, 1,
+                                &pEvdev->emulateMB.enabled,
+                                FALSE);
+    if (rc != Success)
+        return;
+    XISetDevicePropertyDeletable(dev, prop_mbemu, FALSE);
+
+    prop_mbtimeout = MakeAtom(EVDEV_PROP_MIDBUTTON_TIMEOUT,
+                              strlen(EVDEV_PROP_MIDBUTTON_TIMEOUT),
+                              TRUE);
+    rc = XIChangeDeviceProperty(dev, prop_mbtimeout, XA_INTEGER, 32, PropModeReplace, 1,
+                                &pEvdev->emulateMB.timeout, FALSE);
+
+    if (rc != Success)
+        return;
+    XISetDevicePropertyDeletable(dev, prop_mbtimeout, FALSE);
+
+    XIRegisterPropertyHandler(dev, EvdevMBEmuSetProperty, NULL, NULL);
+}
+#endif
